@@ -3,7 +3,7 @@ import collections
 from distutils.version import StrictVersion
 
 import django
-from django.db import models, transaction, OperationalError
+from django.db import models, OperationalError
 from django.db.models.manager import BaseManager
 from django.db.models.base import ModelBase
 from django.utils.translation import ugettext_lazy as _
@@ -14,6 +14,7 @@ from chamber.exceptions import PersistenceException
 from chamber.patch import Options
 from chamber.shortcuts import change_and_save, change, bulk_change_and_save
 from chamber.utils.decorators import singleton
+from chamber.utils.transaction import atomic_with_signals, transaction_signals
 
 from .fields import *  # NOQA exposing classes and functions as a module API
 from .signals import dispatcher_post_save, dispatcher_pre_save
@@ -409,8 +410,12 @@ class SmartModel(AuditModel, metaclass=SmartModelBase):
             self.full_clean(exclude=exclude)
         except ValidationError as er:
             if hasattr(er, 'error_dict'):
-                raise PersistenceException(', '.join(
-                    ('%s: %s' % (key, ', '.join(map(force_text, val))) for key, val in er.message_dict.items())))
+                raise PersistenceException(
+                    ', '.join(
+                        ('%s: %s' % (key, ', '.join(map(force_text, val))) for key, val in er.message_dict.items())
+                    ),
+                    error_dict=er.message_dict
+                )
             else:
                 raise PersistenceException(', '.join(map(force_text, er.messages)))
 
@@ -504,10 +509,11 @@ class SmartModel(AuditModel, metaclass=SmartModelBase):
 
     def save(self, update_only_changed_fields=False, *args, **kwargs):
         if self._smart_meta.is_save_atomic:
-            with transaction.atomic():
+            with atomic_with_signals():
                 self._save(update_only_changed_fields=update_only_changed_fields, *args, **kwargs)
         else:
-            self._save(update_only_changed_fields=update_only_changed_fields, *args, **kwargs)
+            with transaction_signals():
+                self._save(update_only_changed_fields=update_only_changed_fields, *args, **kwargs)
 
     def _pre_delete(self, *args, **kwargs):
         pass
@@ -537,7 +543,7 @@ class SmartModel(AuditModel, metaclass=SmartModelBase):
 
     def delete(self, *args, **kwargs):
         if self._smart_meta.is_delete_atomic:
-            with transaction.atomic():
+            with atomic_with_signals():
                 self._delete(*args, **kwargs)
         else:
             self._delete(*args, **kwargs)
